
#include <zlib.h>
#include "dxa_player.h"
#include "systemstub.h"


bool DXA_Player::open(const char *filename) {
	bool opened = false;
	printf("opening '%s'\n", filename);
	if (_fd.open(filename, "rb")) {
		uint32 tag = _fd.readUint32BE();
		if (tag == _tagDEXA) {
			uint8 flags = _fd.readByte();
			_framesCount = _fd.readUint16BE();
			_frameTicks = _fd.readUint32BE();
			if (_frameTicks > 100) {
				_frameTicks = 100;
			}
			_width = _fd.readUint16BE();
			_height = _fd.readUint16BE();

            if (flags & 0x80)
                _scaleMode = S_INTERLACED;
            else if (flags & 0x40)
                _scaleMode = S_DOUBLE;
            else
                _scaleMode = S_NONE;
                
            if (_scaleMode != S_NONE)
                _workheight = _height / 2;
            else
                _workheight = _height;
                
			printf("frames_count %d width %d height %d ticks %d\n", _framesCount, _width, _height, _frameTicks);
			printf("scaleMode: ");
            switch (_scaleMode) {
                case S_INTERLACED:
                    printf("interlaced\n"); break;
                case S_DOUBLE:
                    printf("double-height\n"); break;
                case S_NONE:
                    printf("none\n"); break;
            }

			_frameSize = _width * _height;
			_frameBuffer1 = (uint8 *)malloc(_frameSize);
			_frameBuffer2 = (uint8 *)malloc(_frameSize);
			_scaledBuffer = (uint8 *)malloc(_frameSize);
			if (!_frameBuffer1 || !_frameBuffer2) {
				printf("error allocating frame tables, size %d\n", _frameSize);
				close();
			} else {
				tag = _fd.readUint32BE();
				if (tag == _tagWAVE) {
					uint32 size = _fd.readUint32BE();
					printf("wave_size = %d\n", size);
					// XXX preload wave data
					_fd.seek(size);
				}
				_currentFrame = 0;
				sysInit(_width, _height, filename);
				opened = true;
			}
		}
	}
	return opened;
}

void DXA_Player::close() {
	sysDestroy();
	_fd.close();
	free(_frameBuffer1);
	free(_frameBuffer2);
	free(_scaledBuffer);
}

void DXA_Player::play() {
	while (_currentFrame < _framesCount) {
		handleNextFrame();
		++_currentFrame;
	}
}

void DXA_Player::handleNextFrame() {
	uint32 tag = _fd.readUint32BE();
	if (tag == _tagCMAP) {
		uint8 pal[256 * 3];
		_fd.read(pal, ARRAYSIZE(pal));
		sysSetPalette(pal, 256);
	}
	tag = _fd.readUint32BE();
	uint8 type;
	if (tag == _tagFRAM) {
		type = _fd.readByte();
		uint32 size = _fd.readUint32BE();
		printf("frame %d type %d size %d\n", _currentFrame, type, size);
		_fd.read(_frameBuffer2, size);
		switch (type) {
		case 2:
		case 3:
			decodeZlib(_frameBuffer2, size, _frameSize);
			break;
		case 4:
		case 5:
			decode0(_frameBuffer2, size);
			break;
		case 6:
		case 7:
			decode2(_frameBuffer2, size, _frameSize);
			break;
		case 13:
			decode13(_frameBuffer2, size, _frameSize);
			break;
		}
		if (type == 2 || type == 4 || type == 6 || type == 13) {
			memcpy(_frameBuffer1, _frameBuffer2, _frameSize);
		} else {
			for (int j = 0; j < _workheight; ++j) {
				for (int i = 0; i < _width; ++i) {
					const int offs = j * _width + i;
					_frameBuffer1[offs] ^= _frameBuffer2[offs];
				}
			}
		}
	}
	
	if (_scaleMode == S_NONE) {
        sysCopyRect(_frameBuffer1, _width, 0, 0, _width, _height);
    } else {
        switch (_scaleMode) {
            case S_INTERLACED:
                memset(_scaledBuffer, 0, _width * _height);
                for (int y = 0; y < _workheight; y++)
                    memcpy(&_scaledBuffer[2*y*_width], &_frameBuffer1[y*_width], _width);
                break;
            case S_DOUBLE:
                for (int y = 0; y < _workheight; y++) {
                    memcpy(&_scaledBuffer[2*y*_width], &_frameBuffer1[y*_width], _width);
                    memcpy(&_scaledBuffer[((2*y)+1)*_width], &_frameBuffer1[y*_width], _width);
                }
                break;
        }
        sysCopyRect(_scaledBuffer, _width, 0, 0, _width, _height);
    }
    
    //if (type == 13) sysDelay(1000);

	if (sysDelay(30)) exit(0);
}

void DXA_Player::decodeZlib(uint8 *data, int size, int totalSize) {
	uint8 *temp = (uint8 *)malloc(size);
	if (temp) {
		memcpy(temp, data, size);
    	z_stream d_stream;
    	d_stream.zalloc = (alloc_func)0;
    	d_stream.zfree = (free_func)0;
    	d_stream.opaque = (voidpf)0;
    	d_stream.next_in = temp;
    	d_stream.avail_in = size;
    	d_stream.total_in = size;
    	d_stream.next_out = data;
    	d_stream.avail_out = totalSize;
    	inflateInit(&d_stream);
        inflate(&d_stream, Z_FINISH);
    	inflateEnd(&d_stream);
		free(temp);
	}
}

void DXA_Player::decode0(uint8 *data, int size) {
	// XXX
}

void DXA_Player::decode2(uint8 *data, int size, int totalSize) {
	// XXX
}

void DXA_Player::decode13(uint8 *data, int size, int totalSize) {
	// XXX

	#define BLOCKW 4
	#define BLOCKH 4

    uint8 *codeBuf, *dataBuf, *motBuf, *maskBuf;

    printf("---------------------------\n");

	/* decompress the input data */
	decodeZlib(data, size, totalSize);

	uint8 *frame2 = (uint8*)malloc(totalSize);

	memcpy(frame2, _frameBuffer1, totalSize);

    int codeSize = _width * _workheight / 16;
    int dataSize, motSize, maskSize;
    memcpy(&dataSize, data, 4);
    memcpy(&motSize, &data[4], 4);
    memcpy(&maskSize, &data[8], 4);

    printf("codeSize = %08X\n", codeSize);
    printf("dataSize = %08X\n", dataSize);
    printf("motSize = %08X\n", motSize);
    printf("maskSize = %08X\n", maskSize);

    codeBuf = &data[12];
    dataBuf = &codeBuf[codeSize];
    motBuf = &dataBuf[dataSize];
    maskBuf = &motBuf[motSize];

    for (int by = 0; by < _workheight; by += BLOCKH) {
        for (int bx = 0; bx < _width; bx += BLOCKW) {
            uint8 type = *codeBuf++;

            uint8 *b2 = (uint8*)frame2 + bx + by * _width;

            switch (type) {
                case 0:
                    break;

                case 1:
                    {
						unsigned short diffMap;

						diffMap = *(unsigned short*)maskBuf;
						maskBuf += 2;

                        for (int yc = 0; yc < BLOCKH; yc++) {
                            for (int xc = 0; xc < BLOCKW; xc++) {
                                if (diffMap & 0x8000) {
                                    b2[xc] = *dataBuf++;
                                }
                                diffMap <<= 1;
                            }
                            b2 += _width;
                        }

                    }
                    break;

                case 2:
                    {
                        uint8 color = *dataBuf++;
                        for (int yc = 0; yc < BLOCKH; yc++) {
                            for (int xc = 0; xc < BLOCKW; xc++) {
                                b2[xc] = color;
                            }
                            b2 += _width;
                        }
                    }
                    break;

                case 3:
                    {
                        for (int yc = 0; yc < BLOCKH; yc++) {
                            for (int xc = 0; xc < BLOCKW; xc++) {
                                b2[xc] = *dataBuf++;
                            }
                            b2 += _width;
                        }
                    }
                    break;

                case 4:
                        {
                                uint8 mbyte = *motBuf++;
                                int mx = (mbyte >> 4) & 0x07;
                                if (mbyte & 0x80) mx = -mx;
                                int my = mbyte & 0x07;
                                if (mbyte & 0x08) my = -my;
								uint8 *b1 = (uint8*)_frameBuffer1 + (bx+mx) + (by+my) * _width;
                                for (int yc = 0; yc < BLOCKH; yc++) {
                                        memcpy(b2, b1, BLOCKW);
                                        b1 += _width;
                                        b2 += _width;
                                }
						}
                        break;

                case 8:
                        {

                            static const int subX[4] = {0, 2, 0, 2};
                            static const int subY[4] = {0, 0, 2, 2};

                            uint8 subMask = *maskBuf++;
                            
                            for (int subBlock = 0; subBlock < 4; subBlock++) {
                                int sx = bx + subX[subBlock], sy = by + subY[subBlock];
                                b2 = (uint8*)frame2 + sx + sy * _width;
                                switch (subMask & 0xC0) {
                                    // 00: skip
                                    case 0x00:
                                        break;
                                    // 01: solid color
                                    case 0x40:
                                        {
                                            uint8 subColor = *dataBuf++;
                                            for (int yc = 0; yc < BLOCKH/2; yc++) {
                                                for (int xc = 0; xc < BLOCKW/2; xc++) {
                                                    b2[xc] = subColor;
                                                }
                                                b2 += _width;
                                            }
                                        }
                                        break;
                                    // 02: motion vector
                                    case 0x80:
                                        {
                                            uint8 mbyte = *motBuf++;
                                            int mx = (mbyte >> 4) & 0x07;
                                            if (mbyte & 0x80) mx = -mx;
                                            int my = mbyte & 0x07;
                                            if (mbyte & 0x08) my = -my;
            								uint8 *b1 = (uint8*)_frameBuffer1 + (sx+mx) + (sy+my) * _width;
                                            for (int yc = 0; yc < BLOCKH/2; yc++) {
                                                    memcpy(b2, b1, BLOCKW/2);
                                                    b1 += _width;
                                                    b2 += _width;
                                            }
                                        }
                                        break;
                                    // 03: raw
                                    case 0xC0:
                                        for (int yc = 0; yc < BLOCKH/2; yc++) {
                                            for (int xc = 0; xc < BLOCKW/2; xc++) {
                                                b2[xc] = *dataBuf++;
                                            }
                                            b2 += _width;
                                        }
                                        break;
                                }
                                subMask <<= 2;
                            }
                            
                        }
                        break;

                case 32:
                case 33:
                case 34:
                    {
                        int count = type - 30;
                        uint8 pixels[4];
                        for (int i = 0; i < count; i++)
                            pixels[i] = *dataBuf++;
                            
                        if (count == 2) {
                            unsigned short code = *(unsigned short*)maskBuf;
                            maskBuf += 2;
                            for (int yc = 0; yc < BLOCKH; yc++) {
                                for (int xc = 0; xc < BLOCKW; xc++) {
                                    b2[xc] = pixels[code & 1];
                                    code >>= 1;
                                }
                                b2 += _width;
                            }
                        } else {
                            unsigned long code = *(unsigned long*)maskBuf;
                            maskBuf += 4;
                            for (int yc = 0; yc < BLOCKH; yc++) {
                                for (int xc = 0; xc < BLOCKW; xc++) {
                                    b2[xc] = pixels[code & 3];
                                    code >>= 2;
                                }
                                b2 += _width;
                            }
                        }

                        break;
                    }

                default:
                    printf("unknown type #############\n");

            }
		}
		
		fflush(stdout);
		
	}


	memcpy(data, frame2, totalSize);
	free(frame2);

    printf("decode done\n"); fflush(stdout);

}

const uint32 DXA_Player::_tagDEXA = 0x44455841;

const uint32 DXA_Player::_tagFRAM = 0x4652414D;

const uint32 DXA_Player::_tagWAVE = 0x57415645;

const uint32 DXA_Player::_tagCMAP = 0x434D4150;
